module jp.ku6.tetris.game;

import std.random;
import jp.ku6.windows.game.application;
import jp.ku6.windows.game.graphics;
import jp.ku6.windows.game.input;
import jp.ku6.windows.game.audio;

/++
 + テトリス
 +/
class Tetris
{
	/++
	 + コンストラクタ
	 +/
	this()
	{
		field_ = new ubyte[FIELD_WIDTH * FIELD_HEIGHT];

		task_ = &fieldResetTask;

		random_.seed(unpredictableSeed);


		resource_ = new Surface(new BitmapFile("block.bmp", 0));

		Font font = new Font(`　 　　　＿＿＿_　　
　　　／　　 　 　＼
　 ／　　─　 　 ─＼　
／ 　　 （●） 　（●） ＼　よく考えたら俺いらないな
|　 　　 　 （__人__）　 　 |　　
/　　　　 ∩ノ ⊃　　／
(　 ＼　／ ＿ノ　|　 |
.＼　“　　／＿＿|　 | 　
　　＼ ／＿＿＿ ／ 　 `, "ＭＳ Ｐゴシック", 18);
		aaSurface_ = new Surface(font);
		
		blockPattern_.width = BLOCK_PIXEL;
		blockPattern_.height = BLOCK_PIXEL;
		blockPattern_.surface = resource_;

		aaPattern_.width = font.width;
		aaPattern_.height = font.height;
		aaPattern_.surface = aaSurface_;

		sprite_.pattern = &blockPattern_;
		deleteAnimationSprite_.pattern = &blockPattern_;
		deleteAnimationSprite_.centerX = BLOCK_PIXEL / 2;
		deleteAnimationSprite_.centerY = BLOCK_PIXEL / 2;

		aaSprite_.pattern = &aaPattern_;

		keybord_ = new Keybord;

		bgm_ = new StreamSound(new OggVorbisFile("nc2003.ogg", 0));
		bgm_.play(true, false);

		lineDeleteSound_ = new StaticSound(new WaveFile("nc17885.wav", 0));
	}

	~this()
	{
		delete lineDeleteSound_;
		delete bgm_;
		delete keybord_;
		delete aaSurface_;
		delete resource_;
	}

	/++
	 + 毎フレームの処理
	 +/
	void mainLoop()
	{
		// キーボード入力状態を更新
		keybord_.refresh();

		// 現在の処理
		task_();

		// 描画処理
		draw();
	}

private:

	// ブロックの種類
	enum Block : ubyte
	{
		NONE = 0,
		WALL = 0x0f,
		DELETING_FLAG = 0x80,
	}

	// ブロックのレイアウト(7種)
	static
	ubyte[] BLOCK_LAYOUTS =
	[
		0,0,0,0,
		0,0,0,0,
		1,1,1,1,
		0,0,0,0,

		0,0,0,0,
		0,0,2,0,
		0,2,2,2,
		0,0,0,0,

		0,0,0,0,
		0,3,3,0,
		0,0,3,3,
		0,0,0,0,

		0,0,0,0,
		0,0,4,4,
		0,4,4,0,
		0,0,0,0,

		0,0,0,0,
		0,0,0,5,
		0,5,5,5,
		0,0,0,0,

		0,0,0,0,
		0,6,0,0,
		0,6,6,6,
		0,0,0,0,

		0,0,0,0,
		0,7,7,0,
		0,7,7,0,
		0,0,0,0,
	];

	// ブロック1個の縦横ピクセル数
	immutable uint BLOCK_PIXEL = 16;

	// フィールドの1行ブロック数
	immutable uint FIELD_WIDTH = 14;

	// フィールドの1列ブロック数
	immutable uint FIELD_HEIGHT = 26;

	// フィールド
	ubyte[] field_;

	// 落下ブロックの左上点のフィールド上座標
	uint blockX_;
	uint blockY_;

	// 落下ブロックのレイアウト
	ubyte[16] block_;

	// ブロックが1段下がるまでのフレーム数
	uint blockDownWait_ = 30;

	// ブロックが1段下がる時を計るカウンタ
	uint blockDownFrame_ = 0;

	// 作業中のタスク
	void delegate() task_;

	// ランダムオブジェクト
	Random random_;

	// ブロックサーフェス
	Surface resource_;

	// AA用サーフェス
	Surface aaSurface_;

	// ブロックサーフェス上のブロック矩形範囲
	Pattern blockPattern_;

	// AA用パターン
	Pattern aaPattern_;

	// ブロック描画用スプライト
	Sprite sprite_;

	// AA描画用スプライト
	Sprite aaSprite_;

	uint deleteAnimationCounter_;

	// 削除アニメーション用のスプライト
	Sprite deleteAnimationSprite_;

	// キーボードオブジェクト
	Keybord keybord_;

	// BGM
	Sound bgm_;

	// 行消去時のSE
	Sound lineDeleteSound_;

	/+
	 + 描画処理
	 +/
	void draw()
	{
		Graphics.instance.begin();

		aaSprite_.color -= 2;
		aaSprite_.color |= 0xff000000;
		aaSprite_.draw();

		float originX = (640 - (FIELD_WIDTH * BLOCK_PIXEL)) / 2;
		float originY = (480 - (FIELD_HEIGHT * BLOCK_PIXEL)) / 2;

		// フィールドを描画
		for (uint y; y < FIELD_HEIGHT; ++y)
		{
			for (uint x; x < FIELD_WIDTH; ++x)
			{
				ubyte type = field_[y * FIELD_WIDTH + x];
				if (type != Block.NONE && (type & Block.DELETING_FLAG) == 0)
				{
					sprite_.x = x * BLOCK_PIXEL + originX;
					sprite_.y = y * BLOCK_PIXEL + originY;
					sprite_.color = typeToColor(type);
					sprite_.draw();
				}
			}
		}

		for (uint y; y < 4; ++y)
		{
			for (uint x; x < 4; ++x)
			{
				ubyte type = block_[y * 4 + x];
				if (type != Block.NONE)
				{
					sprite_.x = (x + blockX_) * BLOCK_PIXEL + originX;
					sprite_.y = (y + blockY_) * BLOCK_PIXEL + originY;
					sprite_.color = typeToColor(type);
					sprite_.draw();
				}
			}
		}

		// 削除アニメーションの描画
		for (uint y; y < FIELD_HEIGHT; ++y)
		{
			for (uint x; x < FIELD_WIDTH; ++x)
			{
				ubyte type = field_[y * FIELD_WIDTH + x];
				if (type != Block.NONE && (type & Block.DELETING_FLAG) != 0)
				{
					deleteAnimationSprite_.x = x * BLOCK_PIXEL + originX + BLOCK_PIXEL / 2;
					deleteAnimationSprite_.y = y * BLOCK_PIXEL + originY + BLOCK_PIXEL / 2;
					deleteAnimationSprite_.color = (deleteAnimationSprite_.color & 0xff000000) | (typeToColor(type ^ Block.DELETING_FLAG) & 0x00ffffff);
					deleteAnimationSprite_.draw();
				}
			}
		}

		Graphics.instance.end();
	}

	/+
	 + フィールドの初期化
	 +/
	void fieldResetTask()
	{
		// 全体を空にする
		for (uint i; i < field_.length; ++i)
			field_[i] = Block.NONE;

		// 底を作る
		for (uint i; i < FIELD_WIDTH; ++i)
			field_[FIELD_WIDTH * (FIELD_HEIGHT - 1) + i] = Block.WALL;

		// 両サイドの壁を作る
		for (uint i = 3; i < FIELD_HEIGHT; ++i)
		{
			field_[FIELD_WIDTH * i] = Block.WALL;
			field_[FIELD_WIDTH * i + FIELD_WIDTH - 1] = Block.WALL;
		}

		// ブロック生成タスクへ
		task_ = &newBlockSetTask;
	}

	/+
	 + 新たにブロックを落下させる
	 +/
	void newBlockSetTask()
	{
		blockX_ = FIELD_WIDTH / 2 - 2;
		blockY_ = 0;

		uint rand = uniform(0, 7);
		for (uint i; i < block_.length; ++i)
			block_[i] = BLOCK_LAYOUTS[rand * 16 + i];

		rand = uniform(0, 4);
		for (uint i; i < rand; ++i)
			rotateBlock(true);


		if (blockIsHitField())
			task_ = &fieldResetTask;
		else
			task_ = &blockDownTask;
	}

	/++
	 + 落下中のブロックを一段落とす
	 +
	 + 落とそうとした時に接触したらそのままフィールドにセットする
	 +/
	void blockDownTask()
	{
		if (keybord_.isPush(Keybord.Key.LEFT))
		{
			--blockX_;
			if (blockIsHitField())
				++blockX_;
		}
		if (keybord_.isPush(Keybord.Key.RIGHT))
		{
			++blockX_;
			if (blockIsHitField())
				--blockX_;
		}
		if (keybord_.isPush(Keybord.Key.Z))
		{
			rotateBlock(true);
			if (blockIsHitField())
				rotateBlock(false);
		}
		if (keybord_.isPush(Keybord.Key.X))
		{
			rotateBlock(false);
			if (blockIsHitField())
				rotateBlock(true);
		}
		if (++blockDownFrame_ >= blockDownWait_ || keybord_.isPress(Keybord.Key.DOWN))
		{
			blockDownFrame_ = 0;

			++blockY_;
			if (blockIsHitField())
			{
				--blockY_;
				for (uint y; y < 4; ++y)
				{
					for (uint x; x < 4; ++x)
					{
						if (block_[y * 4 + x] != Block.NONE)
						{
							field_[(y + blockY_)* FIELD_WIDTH + x + blockX_] = block_[y * 4 + x];
						}
					}
				}

				task_ = &blockDeleteTask;
			}
		}
	}

	/+
	 + 一行揃った箇所を削除する
	 +/
	void blockDeleteTask()
	{
		bool deleteAnimation = false;
		for (uint y; y < FIELD_HEIGHT - 1; ++y)
		{
			bool del = true;
			for (uint x; x < FIELD_WIDTH; ++x)
			{
				if (field_[y * FIELD_WIDTH + x] == Block.NONE)
				{
					del = false;
					break;
				}
			}

			if (del)
			{
				// 該当行をアニメーションにマークする
				for (uint x = 1; x < FIELD_WIDTH - 1; ++x)
					field_[y * FIELD_WIDTH + x] |= Block.DELETING_FLAG;

				deleteAnimation = true;
			}
		}

		if (deleteAnimation)
		{
			deleteAnimationCounter_ = 0;
			deleteAnimationSprite_.rotation = 0;
			deleteAnimationSprite_.scaleX = 1.0;
			deleteAnimationSprite_.scaleY = 1.0;
			block_[0 .. $] = Block.NONE;
			task_ = &lineDeleteAnimationTask;
		}
		else
			task_ = &newBlockSetTask;
	}

	void lineDeleteAnimationTask()
	{
		if (deleteAnimationCounter_ == 0)
		{
			lineDeleteSound_.play(false, true);
		}
		else if (deleteAnimationCounter_ < 30)
		{
			deleteAnimationSprite_.scaleX += 0.05;
			deleteAnimationSprite_.scaleY += 0.05;
			deleteAnimationSprite_.rotation += 32;
			deleteAnimationSprite_.color = 0xff000000;
		}
		else if (deleteAnimationCounter_ < 210)
		{
			deleteAnimationSprite_.scaleX -= 0.01;
			deleteAnimationSprite_.scaleY -= 0.01;
			deleteAnimationSprite_.color = (210 - deleteAnimationCounter_) << 24;
			deleteAnimationSprite_.rotation += 8;
		}
		else
		{
			for (uint y; y < FIELD_HEIGHT - 1; ++y)
			{
				bool del = true;
				for (uint x; x < FIELD_WIDTH; ++x)
				{
					if (field_[y * FIELD_WIDTH + x] == Block.NONE)
					{
						del = false;
						break;
					}
				}

				if (del)
				{
					// その行より上の行全てを一段下に落とす
					for (uint y2 = y; y2 > 0; --y2)
					{
						for (uint x = 1; x < FIELD_WIDTH - 1; ++x)
						{
							field_[y2 * FIELD_WIDTH + x] = field_[(y2 - 1) * FIELD_WIDTH + x];
						}
					}

					// 一番上の行をクリアする
					for (uint x = 1; x < FIELD_WIDTH - 1; ++x)
					{
						field_[x] = Block.NONE;
					}
				}
			}

			task_ = &newBlockSetTask;
		}
		++deleteAnimationCounter_;
	}

	/+
	 + 落下中のブロックがフィールド上のブロックに接触しているか判定
	 +/
	bool blockIsHitField()
	{
		for (uint y; y < 4; ++y)
		{
			for (uint x; x < 4; ++x)
			{
				if (block_[y * 4 + x] != Block.NONE && field_[(y + blockY_)* FIELD_WIDTH + x + blockX_] != Block.NONE)
					return true;
			}
		}

		return false;
	}

	/+
	 + 落下中のブロックを回転
	 +
	 + Params:
	 +  left = true=左回転  false=右回転
	 +/
	void rotateBlock(bool left)
	{
		// 現在の状態を一旦退避する
		ubyte[16] block;
		for (uint i; i < block.length; ++i)
			block[i] = block_[i];

		if (left)
		{
			// 左回転
			block_[0 * 4 + 0] = block[0 * 4 + 3];
			block_[0 * 4 + 1] = block[1 * 4 + 3];
			block_[0 * 4 + 2] = block[2 * 4 + 3];
			block_[0 * 4 + 3] = block[3 * 4 + 3];
			block_[1 * 4 + 0] = block[0 * 4 + 2];
			block_[1 * 4 + 1] = block[1 * 4 + 2];
			block_[1 * 4 + 2] = block[2 * 4 + 2];
			block_[1 * 4 + 3] = block[3 * 4 + 2];
			block_[2 * 4 + 0] = block[0 * 4 + 1];
			block_[2 * 4 + 1] = block[1 * 4 + 1];
			block_[2 * 4 + 2] = block[2 * 4 + 1];
			block_[2 * 4 + 3] = block[3 * 4 + 1];
			block_[3 * 4 + 0] = block[0 * 4 + 0];
			block_[3 * 4 + 1] = block[1 * 4 + 0];
			block_[3 * 4 + 2] = block[2 * 4 + 0];
			block_[3 * 4 + 3] = block[3 * 4 + 0];
		}
		else
		{
			// 右回転
			block_[0 * 4 + 0] = block[3 * 4 + 0];
			block_[0 * 4 + 1] = block[2 * 4 + 0];
			block_[0 * 4 + 2] = block[1 * 4 + 0];
			block_[0 * 4 + 3] = block[0 * 4 + 0];
			block_[1 * 4 + 0] = block[3 * 4 + 1];
			block_[1 * 4 + 1] = block[2 * 4 + 1];
			block_[1 * 4 + 2] = block[1 * 4 + 1];
			block_[1 * 4 + 3] = block[0 * 4 + 1];
			block_[2 * 4 + 0] = block[3 * 4 + 2];
			block_[2 * 4 + 1] = block[2 * 4 + 2];
			block_[2 * 4 + 2] = block[1 * 4 + 2];
			block_[2 * 4 + 3] = block[0 * 4 + 2];
			block_[3 * 4 + 0] = block[3 * 4 + 3];
			block_[3 * 4 + 1] = block[2 * 4 + 3];
			block_[3 * 4 + 2] = block[1 * 4 + 3];
			block_[3 * 4 + 3] = block[0 * 4 + 3];
		}

		// 回転して壁に接触するなら回転させない
		if (blockIsHitField())
		{
			for (uint i; i < block.length; ++i)
				block_[i] = block[i];
		}
	}

	/+
	 + ブロックの種類毎に色を取得
	 +/
	uint typeToColor(ubyte type)
	{
		switch (type)
		{
		case 1: return 0xffffffff;
		case 2: return 0xff80ffff;
		case 3: return 0xffff80ff;
		case 4: return 0xffffff80;
		case 5: return 0xff8080ff;
		case 6: return 0xffff8080;
		case 7: return 0xff80ff80;
		default: return 0xff808080;
		}
	}
}




